"""Various type/class definitions for the Workspaces bit of the Toggle API
Parse/Coercion done by Pydantic
"""

import logging
from datetime import datetime
from typing import Any, Optional

try:
    # Pydantic v2 ships a copy of v1.
    from pydantic.v1 import BaseModel, Field, SecretStr
except ImportError:
    # Home Assistant does not yet support v2.
    from pydantic import BaseModel, Field, SecretStr

from .const import BASE

log = logging.getLogger(__name__)

ENDPOINT = f"{BASE}/workspaces"


class CSVUpload(BaseModel):  # pyright: ignore[reportGeneralTypeIssues]
    """_summary_"""

    at: datetime
    log_id: int


class Subscription(BaseModel):  # pyright: ignore[reportGeneralTypeIssues]
    """Machine generated class representing the Toggl Subscription object."""

    auto_renew: bool
    # Docs say type is: models.CardDetails
    # But not sure where that's actually defined / I don't have an example of it.
    card_details: Any
    company_id: int
    # Docs say type is: models.ContactDetail
    # But not sure where that's actually defined / I don't have an example of it.
    contact_detail: Any
    created_at: str
    currency: str
    customer_id: int
    deleted_at: datetime
    last_pricing_plan_id: int
    organization_id: int
    # Docs say type is: models.PaymentDetails
    # But not sure where that's actually defined / I don't have an example of it.
    payment_details: Any
    pricing_plan_id: int
    renewal_at: datetime
    subscription_id: int

    subscription_period: Any
    workspace_id: int


class Workspace(BaseModel):  # pyright: ignore[reportGeneralTypeIssues]
    """Class representing the Toggl Workspace object.
    Leverages dataclass to cut down on boilerplate code.
    See: https://developers.track.toggl.com/docs/api/workspaces#200
    """

    admin: bool = Field(
        default=False,
        description="Indicates if current user is an Admin for workspace ID (unconfirmed)",
    )

    # Unclear why the server sends this back... It's sensitive so don't print it or include it in model serialization
    # When user instantiates, this could be known but should not need to be set
    api_token: Optional[SecretStr] = Field(exclude=True, repr=False)

    # Seems to be the datetime server side that request was received
    at: Optional[datetime] = Field(
        exclude=True,
        repr=False,
        default=None,
        description="When was last updated",
    )

    business_ws: bool = Field(
        default=None,
        description="Indicates if the workspace is personal or business? (unconfirmed)",
    )

    csv_upload: Optional[CSVUpload] = Field(
        default=None,
        description="TODO",
    )

    default_currency: Optional[str] = Field(
        default=None,
        description="String representation of currency for billable tasks/time entries? (unconfirmed)",
    )

    default_hourly_rate: Optional[float] = Field(
        description="TODO (unconfirmed)", default=None
    )

    ical_enabled: Optional[bool] = Field(
        default=None,
        description="TODO (unconfirmed)",
    )

    ical_url: Optional[str] = Field(default=None, description="TODO (unconfirmed)")

    # Optional because we won't know
    id: Optional[int] = Field(default=None, description="TODO (unconfirmed)")

    last_modified: Optional[datetime] = Field(
        description="TODO (unconfirmed)", default=None
    )

    logo_url: Optional[str] = Field(default=None, description="TODO (unconfirmed)")

    # How far back free workspaces can access data.
    # Docs indicate that it's a thing ... I don't have it in the payloads that I'm getting back
    # Disable for now.
    # max_data_retention_days: Optional[int | None]

    name: str = Field(default=None, description="TODO (unconfirmed)")

    only_admins_may_create_projects: bool = Field(
        default=None,
        description="Indicates if users with Admin roles  (unconfirmed)",
    )

    only_admins_may_create_tags: bool = Field(
        default=None,
        description="TODO (unconfirmed)",
    )

    only_admins_see_billable_rates: bool = Field(
        default=None,
        description="TODO (unconfirmed)",
    )

    only_admins_see_team_dashboard: bool = Field(
        default=None,
        description="TODO (unconfirmed)",
    )

    organization_id: Optional[int] = Field(
        default=None,
        description="ID of the organization that owns this workspace (unconfirmed)",
    )

    premium: Optional[bool] = Field(
        default=None,
        description="TODO (unconfirmed)",
    )

    profile: Optional[int] = Field(
        default=None,
        description="TODO (unconfirmed)",
    )

    projects_billable_by_default: Optional[bool] = Field(
        default=None,
        description="TODO (unconfirmed)",
    )

    rate_last_updated: Optional[datetime] = Field(
        default=None,
        description="TODO (unconfirmed)",
    )

    reports_collapse: Optional[bool] = Field(
        default=None,
        description="TODO (unconfirmed)",
    )

    role: Optional[str] = Field(
        default=None,
        description="TODO (unconfirmed)",
    )

    rounding_minutes: int = Field(
        default=None,
        description="TODO (unconfirmed)",
    )

    # Should be a bool?
    rounding: int = Field(
        default=None,
        description="TODO (unconfirmed)",
    )

    server_deleted_at: Optional[datetime] = Field(
        exclude=True, default=None, description="When was deleted, null if not deleted"
    )

    subscription: Optional[Subscription] = Field(
        default=None,
        description="TODO (unconfirmed)",
    )

    suspended_at: Optional[datetime] = Field(
        exclude=True, default=None, description="When was deleted, null if not deleted"
    )

    working_hours_in_minutes: Optional[int] = Field(
        default=None,
        description="TODO (unconfirmed)",
    )

    # TODO: tags seem to belong to a workspace, so move the get_tags() method here?



"""
lib-toggl
"""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("lib-toggl")
except PackageNotFoundError:
    __version__ = None


__all__ = ["client", "account", "time_entries"]



"""Various type/class definitions for the Organizations bit of the Toggle API
Parse/Coercion done by Pydantic
"""

import logging
from datetime import datetime
from typing import Any, List, Optional

try:
    # Pydantic v2 ships a copy of v1.
    from pydantic.v1 import BaseModel, Field
except ImportError:
    # Home Assistant does not yet support v2.
    from pydantic import BaseModel, Field

from pyrfc3339 import generate

from .const import BASE, DEFAULT_CREATED_BY

log = logging.getLogger(__name__)

ENDPOINT = f"{BASE}/me/time_entries"


@staticmethod
def validate_time_entry_id(time_entry_id: Any) -> None:
    """Raises Value Error if time_entry_id is not a positive integer.
    Allow for `Any` as the type to allow for None to be passed in as a value
    """
    if not time_entry_id:
        raise ValueError("time_entry_id must be specified")
    if not isinstance(time_entry_id, int):
        raise TypeError("time_entry_id must be an integer")
    if time_entry_id <= 0:
        raise ValueError("time_entry_id must be positive.")


@staticmethod
def validate_workspace_id(workspace_id: Any) -> None:
    """Raises Value Error if workspace_id is not a positive integer
    Allow for `Any` as the type to allow for None to be passed in as a value
    """
    if not workspace_id:
        raise ValueError("workspace_id must be specified")
    if not isinstance(workspace_id, int):
        raise TypeError("workspace_id must be an integer")
    if workspace_id <= 0:
        raise ValueError("workspace_id must be positive.")


@staticmethod
# pylint: disable=invalid-name
def CREATE_ENDPOINT(workspace_id: int) -> str:
    """Returns the endpoint for creating a new time entry in the specified workspace"""
    validate_workspace_id(workspace_id)
    return f"{BASE}/workspaces/{workspace_id}/time_entries"


@staticmethod
# pylint: disable=invalid-name
def STOP_ENDPOINT(workspace_id: int, time_entry_id: int) -> str:
    """Returns the endpoint for creating a new time entry in the specified workspace"""
    validate_workspace_id(workspace_id)
    validate_time_entry_id(time_entry_id)
    return f"{BASE}/workspaces/{workspace_id}/time_entries/{time_entry_id}/stop"


@staticmethod
# pylint: disable=invalid-name
def EDIT_ENDPOINT(workspace_id: int, time_entry_id: int) -> str:
    """Returns the endpoint for editing specific time entry in the specified workspace"""
    validate_workspace_id(workspace_id)
    validate_time_entry_id(time_entry_id)
    return f"{BASE}/workspaces/{workspace_id}/time_entries/{time_entry_id}"


@staticmethod
# pylint: disable=invalid-name
def EXPLICIT_ENDPOINT(time_entry_id: int) -> str:
    """Returns the endpoint for editing specific time entry in the specified workspace"""
    validate_time_entry_id(time_entry_id)
    return f"{BASE}/me/time_entries/{time_entry_id}"


class TimeEntry(BaseModel):  # pyright: ignore[reportGeneralTypeIssues]
    """Class representing the Toggl organization object.
    Leverages dataclass to cut down on boilerplate code.
    See: https://developers.track.toggl.com/docs/api/time_entries#200
    """

    # For Pydantic v1, a sub `config` class holds the custom serialization config.
    # See: https://docs.pydantic.dev/1.10/usage/exporting_models/#json_encoders
    ##

    # pylint: disable=too-few-public-methods
    class Config:
        """_summary_"""

        json_encoders = {
            # All datetimes should be stringified into RFC3339 format
            datetime: lambda dt: generate(dt, utc=True, accept_naive=True),
        }

    # When user creates a TimeEntry, this will not be known; it is set by server when successful CREATE request
    id: Optional[int] = Field(default=None)
    # Toggle API has a few fields that are "legacy" and "should not be used" but are still
    #   returned by the API. We store the new/current/correct field value and set up aliases
    #   so the old field names still work.
    ##
    workspace_id: int = Field(description="Workspace ID, required.", default=None)

    project_id: Optional[int] = Field(description="Project ID, optional.", default=None)

    task_id: Optional[int] = Field(alias="tid", default=None)

    user_id: int = Field(
        default=None,
        description="Time Entry creator ID, if omitted will use the requester user ID",
    )

    billable: bool = Field(
        default=False,
        description="Whether the time entry is marked as billable, optional, default false",
    )

    created_with: str = Field(
        description="Must be provided when creating a time entry and should identify the service/application used to create it",
        # value will NOT be provided when making a "get current" call so we default to None
        default=DEFAULT_CREATED_BY,
    )

    # Toggl API wants everything in RFC3339 format which is just a specific flavor of ISO8601
    # Internally, just store everything as a TZ aware datetime with UTC timezone and only convert to
    #   RFC3339 when we need to send it to the API.
    ##
    # Note that Toggle API **requires** that a start datetime be provided when creating a new Time Entry.
    # No longer going to enforce this / default to now() in model creation as there are some valid use cases for
    #   a model that does not have a start time set.
    # As a convenience, the API client will check for start of None and automatically set to now() during the create() calls.
    start: Optional[datetime] = Field(
        default=None,
        # pylint: disable=line-too-long
        description="Start `datetime` in UTC, required when creating a new Time Entry, optional when updating an existing one.",
    )

    stop: Optional[datetime] = Field(
        default=None,
        # pylint: disable=line-too-long
        description="Stop `datetime` in UTC, can be omitted if it's still running or created with 'duration'. If 'stop' and 'duration' are provided, values must be consistent (start + duration == stop)",
    )

    # Duration in seconds
    # Should be -1 for something that's on-going but should be correct for creating an "already-done" TE
    # e.g. start + duration = stop
    # TODO: better validation around this field with the After validation
    # https://docs.pydantic.dev/2.5/concepts/validators/#before-after-wrap-and-plain-validators
    duration: int = Field(
        default=-1,
        description="Time entry duration. For running entries should be negative, preferable `-1`",
    )

    # Description is required when CREATING, but not required when deleting.
    description: Optional[str] = Field(
        default=None, description="Time entry description, optional"
    )

    tag_action: Optional[str] = Field(pattern=r"^(add|delete)$", default="add")

    tags: Optional[List[str]] = Field(
        default=None,
        description="Tag names",
    )

    tag_ids: Optional[List[int]] = Field(
        default=None,
        description="Tag IDs.",
    )

    # This field is deprecated for GET endpoints where the value will always be true.
    duronly: Optional[bool] = Field(
        exclude=True,
        default=False,
        repr=False,
        description="Deprecated: Used to create a time entry with a duration but without a stop time. This parameter can be ignored.",
    )

    # This appears to be the datetime server got/fulfilled request
    # Isn't something user will supply when creating a Time Entry and doesn't really serve a useful
    #   purpose so we exclude it from the model.
    at: Optional[datetime] = Field(
        exclude=True, default=None, description="When was last updated", repr=False
    )

    server_deleted_at: Optional[datetime] = Field(
        exclude=True, default=None, description="When was deleted, null if not deleted"
    )

    uid: Optional[int] = Field(
        exclude=True,
        default=None,
        repr=False,
        description="Time Entry creator ID, legacy field",
    )

    wid: Optional[int] = Field(
        exclude=True, default=None, repr=False, description="Workspace ID, legacy field"
    )

    pid: Optional[int] = Field(
        exclude=True, default=None, repr=False, description="Project ID, legacy field"
    )

    tid: Optional[int] = Field(
        exclude=True, default=None, repr=False, description="Task ID, legacy field"
    )



"""Tests Pydantic validation around Time Entry object instantiation and modification"""

# pylint: disable=missing-function-docstring

import pytest

from lib_toggl.const import BASE
from lib_toggl.time_entries import (
    CREATE_ENDPOINT,
    EDIT_ENDPOINT,
    EXPLICIT_ENDPOINT,
    STOP_ENDPOINT,
    TimeEntry,
)

##
# Mostly dumb tests to ensure that the endpoints are generated correctly.
##


def test_create_endpoint_with_valid_workspace_id():
    workspace_id = 123
    expected_endpoint = f"{BASE}/workspaces/123/time_entries"
    assert CREATE_ENDPOINT(workspace_id) == expected_endpoint


def test_create_endpoint_with_zero_workspace_id():
    workspace_id = 0
    with pytest.raises(ValueError):
        CREATE_ENDPOINT(workspace_id)


def test_create_endpoint_with_negative_workspace_id():
    workspace_id = -123
    with pytest.raises(ValueError):
        CREATE_ENDPOINT(workspace_id)


def test_create_endpoint_with_none_workspace_id():
    with pytest.raises(TypeError):
        # pylint: disable=no-value-for-parameter
        CREATE_ENDPOINT()  # pyright: ignore reportCallIssue


def test_stop_endpoint_with_valid_ws_te_ids():
    workspace_id = 123
    time_entry_id = 456
    expected_endpoint = (
        f"{BASE}/workspaces/{workspace_id}/time_entries/{time_entry_id}/stop"
    )
    assert STOP_ENDPOINT(workspace_id, time_entry_id) == expected_endpoint


def test_stop_endpoint_with_invalid_ws_te_ids():
    workspace_id = 0
    time_entry_id = -456
    with pytest.raises(ValueError):
        STOP_ENDPOINT(workspace_id, time_entry_id)


def test_stop_endpoint_with_missing_workspace_id():
    with pytest.raises(ValueError):
        STOP_ENDPOINT(None, 456)  # pyright: ignore reportArgumentType


def test_stop_endpoint_with_missing_te_id():
    with pytest.raises(ValueError):
        STOP_ENDPOINT(123, None)  # pyright: ignore reportArgumentType


def test_stop_endpoint_with_none_workspace_id():
    with pytest.raises(TypeError):
        # pylint: disable=no-value-for-parameter
        STOP_ENDPOINT()  # pyright: ignore reportCallIssue


def test_edit_endpoint_with_valid_workspace_valid_te():
    workspace_id = 123
    time_entry_id = 456
    expected_endpoint = f"{BASE}/workspaces/{workspace_id}/time_entries/{time_entry_id}"
    assert EDIT_ENDPOINT(workspace_id, time_entry_id) == expected_endpoint


def test_edit_endpoint_with_valid_workspace_id_missing_te():
    workspace_id = 123
    with pytest.raises(TypeError):
        # pylint: disable=no-value-for-parameter
        EDIT_ENDPOINT(workspace_id)  # pyright: ignore reportCallIssue


def test_edit_endpoint_with_invalid_workspace_id_valid_te():
    te_id = 456
    with pytest.raises(ValueError):
        EDIT_ENDPOINT(None, te_id)  # pyright: ignore reportCallIssue


def test_edit_endpoint_with_valid_workspace_id_invalid_te():
    ws_id = 123
    with pytest.raises(ValueError):
        EDIT_ENDPOINT(ws_id, None)  # pyright: ignore reportCallIssue


def test_explicit_endpoint_with_valid_te_id():
    time_entry_id = 456
    expected_endpoint = f"{BASE}/me/time_entries/{time_entry_id}"
    assert EXPLICIT_ENDPOINT(time_entry_id) == expected_endpoint


def test_explicit_endpoint_with_zero_te_id():
    time_entry_id = 0
    with pytest.raises(ValueError):
        EXPLICIT_ENDPOINT(time_entry_id)


def test_explicit_endpoint_with_negative_workspace_id():
    time_entry_id = -123
    with pytest.raises(ValueError):
        EXPLICIT_ENDPOINT(time_entry_id)


def test_explicit_endpoint_with_none_workspace_id():
    with pytest.raises(TypeError):
        # pylint: disable=no-value-for-parameter
        EXPLICIT_ENDPOINT()  # pyright: ignore reportCallIssue


##
# Validating TimeEntry object instantiation with sample data from toggl API
# Note: Pydantic is already well tested so there's not a ton of value in testing the basics here.
# Most of the tests below are there to ensure that the additional validation that I added to
#   the base pydantic model is working as expected.
##


def test_time_entry_instantiation_with_valid_data():
    """Aside from the inconsistent behavior and docs, the basic Pydantic/Field() logic is sound.
    The only real thing that I had to add was when the API would return Tags = None, Tag IDs = [] or
    vice versa.
    """
    valid_no_tags = {
        "id": 123,
        "description": "Test Description",
        "tags": None,
        "tag_ids": [],
        "workspace_id": 456,
    }

    valid_te_no_tags_ids = {
        "id": 123,
        "description": "Test Description",
        "tags": [],
        "tag_ids": None,
        "workspace_id": 456,
    }

    valid_no_tags = TimeEntry(**valid_no_tags)
    valid_no_tags_ids = TimeEntry(**valid_te_no_tags_ids)

    assert valid_no_tags.id == 123
    assert valid_no_tags.id == 123

    assert valid_no_tags.workspace_id == 456
    assert valid_no_tags_ids.workspace_id == 456

    assert valid_no_tags.description == "Test Description"
    assert valid_no_tags_ids.description == "Test Description"

    # pylint: disable=use-implicit-booleaness-not-comparison
    assert valid_no_tags.tags is None
    assert valid_no_tags_ids.tags == []

    assert valid_no_tags.tag_ids == []
    assert valid_no_tags_ids.tag_ids is None



"""Account Information
Parse/Coercion done by Pydantic
"""

from datetime import datetime
from typing import Optional

try:
    # Pydantic v2 ships a copy of v1.
    from pydantic.v1 import BaseModel, SecretStr
except ImportError:
    # Home Assistant does not yet support v2.
    from pydantic import BaseModel, SecretStr

from .const import BASE

ENDPOINT = f"{BASE}/me"


class Account(BaseModel):  # pyright: ignore[reportGeneralTypeIssues]
    """
    https://developers.track.toggl.com/docs/api/me#200
    """

    id: int
    api_token: SecretStr
    email: str
    fullname: str
    timezone: str
    toggl_accounts_id: str
    default_workspace_id: int
    # TODO: define weekday type, convert
    beginning_of_week: int
    image_url: str
    created_at: datetime
    updated_at: datetime
    openid_email: Optional[str]
    openid_enabled: bool
    # See: https://github.com/kquinsland/lib-toggl/issues/15
    country_id: Optional[int]
    has_password: bool
    at: datetime
    # Analytics?
    intercom_hash: SecretStr



"""
Very basic Toggl API wrapper
"""

from datetime import UTC, datetime
from typing import Any, List

import aiohttp
from pyrfc3339 import generate


from .account import ENDPOINT as ACCOUNT_ENDPOINT
from .account import Account
from .const import USER_AGENT
from .tags import TAGS_ENDPOINT, Tag
from .time_entries import CREATE_ENDPOINT as TIME_ENTRY_CREATE_ENDPOINT
from .time_entries import EDIT_ENDPOINT as TIME_ENTRY_EDIT_ENDPOINT
from .time_entries import ENDPOINT as TIME_ENTRY_ENDPOINT
from .time_entries import EXPLICIT_ENDPOINT
from .time_entries import STOP_ENDPOINT as TIME_ENTRY_STOP_ENDPOINT
from .time_entries import TimeEntry, validate_time_entry_id, validate_workspace_id
from .workspace import ENDPOINT as WORKSPACE_ENDPOINT
from .workspace import Workspace

# Try structlog (available in dev context), fall back to stdlib logging
try:
    import structlog

    log = structlog.get_logger()

except ImportError:
    import logging

    log = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
class Toggl:
    """Processes the client request

    Raises:
        ValueError: If the request method is not supported.
        ValueError: If the request URL is not valid.
        ValueError: If the request headers are not properly formatted.
        ValueError: If the request body is not properly formatted.
        ValueError: If the request timeout is not a positive number.
        ValueError: If the request fails due to network issues.
        ValueError: If the server responds with an error status code.

    Returns:
        Response: The server's response to the client's request.
    """

    # Basically everything is JSON
    _headers = {"content-type": "application/json"}

    # default API user agent value
    _user_agent = USER_AGENT

    def __init__(self, api_key: str | None) -> None:
        self.headers = {}
        self._session = aiohttp.ClientSession()

        self._account: Account | None = None
        self._current_time_entry: TimeEntry | None = None
        self._workspaces: List[Workspace] | None = None

        self._auth = None

        if api_key:
            self.api_key = api_key
            self._auth = aiohttp.BasicAuth(
                login=api_key, password="api_token", encoding="utf-8"
            )
        else:
            self._api_key = None

    async def __aenter__(self):
        """Alternative: See: https://stackoverflow.com/a/67577364"""
        return self

    async def __aexit__(self, *excinfo):
        await self._session.close()

    async def close(self) -> None:
        """Closes the underlying aiohttp session.

        Needed when not using with X as Y context manager pattern."""
        await self._session.close()

    @property
    def api_key(self) -> str | None:
        """Current API key."""
        return self._api_key

    @api_key.setter
    def api_key(self, value: str):
        """
        Toggl supports Basic Authentication with two flavors: email/pass and api-token.
        Only supports the api-token flavor; oddly enough the api token takes place of the email and
            the password is `api_token`.

        Toggl does not have details on what makes a valid API token.
        Looks like 32 alphanumeric chars but I don't want to code in that assumption.
        """
        if not value:
            raise ValueError("api_key cannot be None")
        self._api_key = value
        self._auth = aiohttp.BasicAuth(
            login=self._api_key, password="api_token", encoding="utf-8"
        )

    @property
    async def account(self) -> Account | None:
        """Toggle Account details."""
        if self._account is None:
            self._account = await self.get_account_details()
        return self._account

    @property
    async def workspaces(self) -> List[Workspace] | None:
        """List of Workspaces the user has access to."""
        if self._workspaces is None:
            self._workspaces = await self.get_workspaces()
        return self._workspaces

    @property
    async def current_time_entry(self) -> TimeEntry | None:
        """Currently running Time Entry, if one exists."""
        if self._current_time_entry is None:
            self._current_time_entry = await self.get_current_time_entry()
        return self._current_time_entry

    async def _pre_flight_check(self):
        """Common pre-request checks"""
        if self._api_key is None:
            raise ValueError("api_key must be set before making requests.")

        if self._session.closed:
            log.error("session is closed, creating new session")
            self._session = aiohttp.ClientSession()

        # Merge common headers with instance specific headers
        self.headers.update(self._headers)

    async def do_get_request(
        self, url: str, data: dict | None = None
    ) -> dict[str, Any]:
        """Does a GET request to the specified URL.

        Args:
            url (str): URL to send the GET request to.
            data (dict | None, optional): A dictionary to be passed as query parameters. Defaults to None.

        Returns:
            Response: The server's response to the GET request.
        """

        await self._pre_flight_check()
        log.debug("do_get_request", extra={"data": data})
        # TODO: error handling, lots of ways network/json parse can fail...
        async with self._session.get(
            url, headers=self.headers, auth=self._auth, params=data
        ) as resp:
            if resp.status != 200:
                log.debug("here is resp", extra={"resp": await resp.text()})
                resp.raise_for_status()
            return await resp.json()

    async def do_post_request(self, url: str, data_as_json_str: str) -> dict[str, Any]:
        """Does a POST request to the specified URL.

        Args:
            url (str): URL to send the POST request to.
            data_as_json_str (str): String encoded JSON data. Not using built-in json kwarg because we need to use the json encoder in pydantic so we can exclude None values.

        Returns:
            Response: The server's response to the POST request.
        """

        await self._pre_flight_check()

        log.debug("do_post_request", extra={"data": data_as_json_str})
        # TODO: error handling, lots of ways network/json parse can fail...
        async with self._session.post(
            url, headers=self.headers, auth=self._auth, data=data_as_json_str
        ) as resp:
            if resp.status != 200:
                log.debug("here is resp", extra={"resp": await resp.text()})
                resp.raise_for_status()
            return await resp.json()

    async def do_patch_request(
        self, url: str, data: dict | None = None
    ) -> dict[str, Any]:
        """Performs a PATCH request to the specified URL.

        Args:
            url (str): URL to send the PATCH request to.
            data (dict | None, optional): A dictionary to be sent as JSON in the body of the request. Defaults to None.

        Returns:
            Response: The server's response to the PATCH request.
        """
        await self._pre_flight_check()

        log.debug("do_patch_request", extra={"data": data})
        # TODO: error handling, lots of ways network/json parse can fail...
        async with self._session.patch(
            url, headers=self.headers, auth=self._auth, data=data
        ) as resp:
            if resp.status != 200:
                log.debug("here is resp", extra={"resp": await resp.text()})
                resp.raise_for_status()
            return await resp.json()

    async def do_put_request(self, url: str, data_as_json_str: str) -> dict[str, Any]:
        """Does PUT request to the specified URL.

        Args:
            url (str): URL to send the PUT request to.
            data (str): JSON encoded string to send in the body of the request.

        Returns:
            dict[str, Any]: JSON response from the server.
        """
        await self._pre_flight_check()

        log.debug("do_put_request", extra={"data": data_as_json_str})
        # TODO: error handling, lots of ways network/json parse can fail...
        async with self._session.put(
            url, headers=self.headers, auth=self._auth, data=data_as_json_str
        ) as resp:
            if resp.status != 200:
                log.debug("here is resp", extra={"resp": await resp.text()})
                resp.raise_for_status()

            return await resp.json()

    ##
    # Actual methods for fetching things from Toggl
    ##

    async def get_workspaces(self) -> List[Workspace]:
        """Gets a list of Workspaces the user has access to.

        Returns:
            [Workspace]: List of Workspace objects.
        """
        log.debug("get_workspaces is alive...")
        ws = await self.do_get_request(WORKSPACE_ENDPOINT)
        log.debug("get_workspaces", extra={"ws": ws})
        # As of now, not a ton of error handling in the do_*_request functions.
        # We do basic checking here to make sure pylance is happy.
        if ws is None:
            log.debug("No workspaces found")
            return []
        # Assuming nothing went wrong, `ws` will be a list with one json object per workspace
        return [Workspace(**x) for x in ws]  # pyright: ignore reportCallIssue

    async def get_tags(self, workspace_id: int) -> List[Tag]:
        """Returns a list of Tags for the specified workspace.

        Args:
            workspace_id (int): Workspace ID to fetch tags for.

        Returns:
            List[Tag]: List of Tag objects.
        """
        tags = await self.do_get_request(TAGS_ENDPOINT(workspace_id))
        log.debug("get_tags", extra={"tags": tags})
        # As of now, not a ton of error handling in the do_*_request functions.
        # We do basic checking here to make sure pylance is happy.
        if tags is None:
            log.debug("No workspaces found")
            return []
        # Assuming nothing went wrong, `tags` will be a list with one json object per tag
        return [Tag(**x) for x in tags]  # pyright: ignore reportCallIssue

    async def create_tag(self, workspace_id: int, tag_name: str) -> Tag | None:
        """Creates a new Tag in the specified workspace.

        Args:
            workspace_id (int): Workspace ID to create the tag in.
            tag_name (str): Name of the tag to create.

        Returns:
            Tag | None: Tag object if successful, else None.
        """
        body = {"name": tag_name, "workspace_id": workspace_id}
        _t = Tag(**body)
        data = _t.json(exclude_none=True)
        log.debug("create_tag. To make: %s", data)
        d = await self.do_post_request(
            TAGS_ENDPOINT(workspace_id), data_as_json_str=data
        )
        # As of now, not a ton of error handling in the do_*_request functions.
        # We do basic checking here to make sure pylance is happy.
        if d is None:
            log.debug("Tag not created?")
            return None
        return Tag(**d)

    async def get_time_entries(
        self,
        start_date: datetime,
        end_date: datetime,
    ) -> List[TimeEntry]:
        """Retrieves a list of Time Entries within a specific date range

        Args:
            start_date (datetime): The start date of the range.
            end_date (datetime): The end date of the range.

        Raises:
            ValueError: If the start_date is later than the end_date.
            ValueError: If either the start_date or end_date is not a valid datetime.

        Returns:
            List[TimeEntry]: A list of TimeEntry objects within the specified date range.
        """
        log.info("get_time_entries is alive...")

        # Start and end date must be provided together
        if start_date and not end_date:
            raise ValueError("start_date provided but not end_date")
        if end_date and not start_date:
            raise ValueError("end_date provided but not start_date")

        # Toggle wants RFC3339 formatted strings
        _start = generate(start_date, utc=True, accept_naive=True)
        _end = generate(end_date, utc=True, accept_naive=True)

        # When I don't include these params, things are fine.
        # When
        params = {"start_date": _start, "end_date": _end}

        time_entries = await self.do_get_request(TIME_ENTRY_ENDPOINT, data=params)
        log.debug("get_time_entries", extra={"time_entries": time_entries})
        # As of now, not a ton of error handling in the do_*_request functions.
        # We do basic checking here to make sure pylance is happy.
        if time_entries is None:
            log.debug("No time entries found")
            return []
        # Assuming nothing went wrong, `time_entries` will be a list with one json object per time entry
        return [TimeEntry(**x) for x in time_entries]  # pyright: ignore reportCallIssue

    async def get_current_time_entry(self) -> TimeEntry | None:
        """Returns active Time Entry if one is running, else None"""
        log.info("get_current_time_entry is alive...")

        cte = await self.do_get_request(f"{TIME_ENTRY_ENDPOINT}/current")
        if cte is None:
            log.debug("There doesn't seem to be a currently running Time Entry")
            return None
        try:
            log.debug("get_current_time_entry", extra={"cte": cte})
            self._current_time_entry = TimeEntry(**cte)
            return await self.current_time_entry

        # pylint: disable-next=broad-except
        except Exception as exc:
            log.debug("err", exc_info=exc)
        return None

    async def get_time_entry_by_id(self, time_entry_id: int) -> TimeEntry | None:
        """Retrieves a specific Time Entry by its ID

        Args:
            time_entry_id (int): The ID of the Time Entry to retrieve.

        Returns:
            TimeEntry | None: The TimeEntry object with the given ID if it exists, else None.
        """
        log.info("get_current_time_entry is alive...")

        te = await self.do_get_request(f"{EXPLICIT_ENDPOINT(time_entry_id)}")
        if te is None:
            log.debug("There doesn't seem to be a currently running Time Entry")
            return None
        try:
            log.debug("get_current_time_entry", extra={"cte": te})
            return TimeEntry(**te)

        # pylint: disable-next=broad-except
        except Exception as exc:
            log.debug("err", exc_info=exc)
        return None

    async def get_account_details(self) -> Account | None:
        """Retrieves the account details

        Returns:
            Account | None: The Account object containing the details of the current account if the operation is successful, else None.
        """
        d = await self.do_get_request(ACCOUNT_ENDPOINT)
        log.debug("get_account_details", extra={"data": d})
        # As of now, not a ton of error handling in the do_*_request functions.
        # We do basic checking here to make sure pylance is happy.
        if d is None:
            return None
        self._account = Account(**d)
        return await self.account

    async def create_new_time_entry(self, te: TimeEntry) -> TimeEntry | None:
        """Creates a new Toggl Track Time Entry

        Args:
            te (TimeEntry): Time Entry object to create. If the `start` property is not set, it will be set to the current time.

        Raises:
            ValueError: If the provided TimeEntry object is not valid.
            ValueError: If the operation on the TimeEntry object fails.

        Returns:
            TimeEntry: The newly created TimeEntry object if the operation is successful, else None.
        """
        log.debug("create_new_time_entry is alive...")
        # This will be the first POST/PUT request
        # Also need to collect / validate data
        # Can leverage pydantic for that, but need to be mindful of that class / what data ONLY comes back with a GET
        # Versus what data should be populated for a PUT
        ##
        _url = TIME_ENTRY_CREATE_ENDPOINT(te.workspace_id)

        if not te.start:
            te.start = datetime.now(UTC)
            log.debug("te.start was not set, setting to %s", te.start)

        # Render out to JSON, exclude the things that user didn't set
        # In testing, it looks like tag_ids will override tags; the list of strings is virtually meaningless!
        # If tags is set to a list of strings but tag_action is not set or tag_ids is an empty list, the server will NOT
        #   set the tags and will create a new time track entry with the empty list of tags.
        # So, we default both tags and tag_ids to None and let the user pick which to set.
        # TODO: i'll want to do more sophisticated validation / coercion to handle this case.
        #   e.g: tag_action should remain None unless tags is a list with at least one string, then default to add
        ##
        data = te.json(exclude_none=True)
        log.debug("create_new_time_entry. To make: %s", data)
        d = await self.do_post_request(_url, data_as_json_str=data)
        # As of now, not a ton of error handling in the do_*_request functions.
        # We do basic checking here to make sure pylance is happy.
        if d is None:
            return None
        return TimeEntry(**d)

    async def _persist_time_entry(self, te: TimeEntry) -> TimeEntry | None:
        """Lower level level API that attempts to update state for an existing Time Entry.
        Can be used directly, is meant to be used by higher level API functions like edit_time_entry().

        Args:
            te (TimeEntry): Object representing the desired state.

        Returns:
            TimeEntry | None: Object representing the persisted state or None on failure.
        """
        # If user creates a TimeEntry directly and tries to use _persist_time_entry instead of create()
        validate_workspace_id(te.workspace_id)
        validate_time_entry_id(te.id)

        _url = TIME_ENTRY_EDIT_ENDPOINT(
            te.workspace_id,
            te.id,  # pyright: ignore reportArgumentType
        )
        data = te.json(exclude_none=True)
        log.debug("_persist_time_entry. sending: %s", data)
        d = await self.do_put_request(_url, data_as_json_str=data)
        # As of now, not a ton of error handling in the do_*_request functions.
        # We do basic checking here to make sure pylance is happy.
        if d is None:
            return None
        return TimeEntry(**d)

    async def edit_time_entry(self, local_te: TimeEntry) -> TimeEntry | None:
        """High level API that attempts to update state for an existing Time Entry.

        Modifying some aspects of a Time Entry are trivial, such as the description.
        Tags are a bit more convoluted. Depending on the passed time entry and the desired state
            several API calls may be needed.
        This is abstracted away in the update_tags() function which is called by this function.
        That function is meant to be relatively cheap to call if the user ends up passing in a desired set of tags
            that perfectly overlaps with the current set of tags on the Time Entry.
        In that case, we'll just move on to updating the description for the Time Entry

        Args:
            local_te (TimeEntry): Object representing the desired state.

        Returns:
            TimeEntry | None: Object representing the persisted state or None on failure.
        """
        log.debug("edit_time_entry is alive. Starting with %s", local_te)

        # To determine if we have any tags to update, we first need to ask the server what IT thinks the passed in TE looks like
        remote_te = await self.get_time_entry_by_id(
            local_te.id  # pyright: ignore reportArgumentType
        )
        log.debug("remote_te: %s", remote_te)
        if remote_te is None:
            log.error(
                "Failed to fetch remote Time Entry. Can't edit what server doesn't know about."
            )
            return None

        # update_tags() requires non-None inputs
        if local_te.tags is None:
            local_te.tags = []
        # Do tag updates, if any.
        updated_te = await self.update_tags(remote_te, local_te.tags)
        log.debug("edit_time_entry: updated_te: %s", updated_te)
        if updated_te is None:
            log.error(
                "Failed to update tags on Time Entry, attempting to update other fields, though"
            )
            updated_te = local_te
        # Update the description to match what the user passed in
        updated_te.description = local_te.description
        log.debug("edit_time_entry: [returning] updated_te: %s", updated_te)
        return await self._persist_time_entry(updated_te)

    async def stop_time_entry(self, te: TimeEntry) -> TimeEntry | None:
        """Summary

        Args:
            te (TimeEntry): The TimeEntry object to be processed.

        Returns:
            TimeEntry | None: The processed TimeEntry object if successful, None otherwise.
        """
        validate_workspace_id(te.workspace_id)
        validate_time_entry_id(te.id)

        log.info(
            "stop_time_entry is alive...",
            extra={"id": te.id, "workspace": te.workspace_id},
        )
        _url = TIME_ENTRY_STOP_ENDPOINT(
            te.workspace_id,
            te.id,  # pyright: ignore reportArgumentType
        )

        d = await self.do_patch_request(_url)
        # As of now, not a ton of error handling in the do_*_request functions.
        # We do basic checking here to make sure pylance is happy.
        if d is None:
            return None
        return TimeEntry(**d)

    async def update_tags(self, te: TimeEntry, new_tags: List[str]) -> TimeEntry | None:
        """
        A wrapper to abstract the logic of updating the tags on a TimeEntry object.

        The Toggl API docs are a bit confusing and ambiguous about how tags are handled when updating a Time Entry.
        As it turns out, the "just give us the string, we'll do the rest" behavior applies only when creating a new Time Entry.

        If you have an existing Time Entry and want to update the tags, you must provide the tag IDs, not the tag strings.

        Furthermore, if you clear and then set the `tag_ids` field on a Time Entry and POST/PUT it to the server, you'd expect
            the time entry to have just the new tags, right? Wrong.
        The tags on the Time Entry will depend on what the `tag_action` was set to when the POST/PUT request was made with
            the tag_ids.

        So instead, we offer a nice simple interface to the user: just give us the list of strings you want the Time Entry to have.
        This means that we need to figure out which tags are new and which already exist in the workspace
        New tags need to be created and then we can get the tag IDs.

        Then we need to figure out which tag_ids should be added and or removed from the Time Entry.

        We take a best effort approach to updating the tags on the Time Entry. Add new tags first, then remove old tags.
        This way, even if there's an error removing tags, the Time Entry will still have the new tags that the user wanted.
        They can always manually search for the old tag(s) and remove them if necessary.
        """
        validate_workspace_id(te.workspace_id)
        validate_time_entry_id(te.id)
        if new_tags is None:
            raise ValueError("new_tags is required.")

        log.debug("Updating tags on TimeEntry: %s.", te.description)
        # TODO: much better error handling
        known_tags = await self.get_tags(te.workspace_id)
        if known_tags is None:
            log.error("Failed to get tags for workspace: %s.", te.workspace_id)
            return

        # Even though get_tags should never return a Tag with either field as None
        # Pylance only sees that Pydantic has either None or Int/Str for the fields
        known_tags = {
            tag.name: tag.id
            for tag in known_tags
            if tag.name is not None and tag.id is not None
        }
        log.debug("Found %s known tags: %s.", len(known_tags), known_tags)

        # Figure out what tags need to be created, added and removed
        tags_to_create = set(new_tags) - set(known_tags.keys())
        log.debug("%s Tags to create: %s", len(tags_to_create), tags_to_create)

        # Pylance only sees that `tags` is Optional[List[str]]. It knows that Pydanitc's Optional means None is possible.
        # It can't see that there is a post validation function that replaces None with [].
        tags_to_add = set(new_tags) - set(te.tags)  # pyright: ignore reportArgumentType
        log.debug("%s Tags to add: %s", len(tags_to_add), tags_to_add)

        # fmt: off
        tags_to_remove = set(te.tags) - set(new_tags) # pyright: ignore reportArgumentType
        log.debug("%s Tags to remove: %s", len(tags_to_remove), tags_to_remove)

        if (
            len(tags_to_create) == 0
            and len(tags_to_add) == 0
            and len(tags_to_remove) == 0
        ):
            log.info("No changes to tags needed.")
            return te

        # Technically, if the TimeEntry that the user passed in has a valid ID but a different description / start / stop / duration... etc
        #   from what the server has handy, calling _persist_time_entry will update the Time Entry to match the user's Time Entry in it's entirety.
        # The name of this function implies that it's only for updating tags so we unset any attributes on the current Time Entry that aren't tags.
        # This way, when we persist the Time Entry, we only update the tags on the server.
        # Easiest way to do this is to clone the Time Entry that user passed in and then unset the non-tag fields; leave the TE they passed in alone.
        ##
        te = TimeEntry(
            id=te.id,
            workspace_id=te.workspace_id,
            tags=te.tags,
            tag_ids=te.tag_ids,
            tag_action=te.tag_action,
        )

        # Create the new tags
        for tag in tags_to_create:
            log.info("Creating new tag: %s", tag)
            new_tag = await self.create_tag(te.workspace_id, tag)
            if new_tag is None:
                log.error("Failed to create tag: %s", tag)
                continue
            if new_tag.name is None or new_tag.id is None:
                log.error(
                    "Tag creation failed: [%s] (%s,%s)", tag, new_tag.id, new_tag.name
                )
                continue
            # Creation was successful, add the new tag to the known tags
            known_tags[new_tag.name] = new_tag.id

        # Assuming nothing went wrong, we should now have an updated list of known tags
        for tag in tags_to_add:
            log.info("Adding tag: %s", tag)
            if tag not in known_tags:
                log.error("Tag not found in known tags: %s. Creation failed?", tag)
                continue
            # Update the Time Entry with the new tag ID and name to be sure.
            # API appears to be inconsistent about weather the string or the ID matters more depending on the action!?!!
            ##
            # pylint: disable=no-member
            te.tag_ids.append(known_tags[tag]) # pyright: ignore reportOptionalMemberAccess
            # pylint: disable=no-member
            te.tags.append(tag) # pyright: ignore reportOptionalMemberAccess

        updated_te = await self._persist_time_entry(te)
        if updated_te is None:
            log.error(
                "Failed to update Time Entry with new tags; refusing to remove old tags (if any)"
            )
            return

        # Assuming nothing broke, the time entry should have tag IDs for new tags and old tags.
        # Clear the tag_ids / strings, set the tag_action to remove, and then set string/ids to the tags that need to be removed.
        if len(tags_to_remove) > 0:
            updated_te.tag_ids = []
            updated_te.tags = []
            updated_te.tag_action = "remove"
            for tag in tags_to_remove:
                log.debug("Removing tag: %s", tag)
                if tag not in known_tags:
                    log.error("Tag not found in known tags: %s. Cannot remove.", tag)
                    continue
                # Again, when remove is the action, behavior is inconsistent when the tag_ids versus tag strings don't match
                updated_te.tag_ids.append(known_tags[tag])
                updated_te.tags.append(tag)

        # Should return a Time Entry with just the new tags
        return await self._persist_time_entry(updated_te)


# TODO: General exceptions to handle and wrap
# Trying to stop a TE that was deleted:
#   aiohttp.client_exceptions.ClientResponseError: 404, message='Not Found',
# Incorrect basic auth
#   aiohttp.client_exceptions.ClientResponseError: 401, message='Unauthorized',
# Trying to stop an entry on a workspace that's not mine
#   aiohttp.client_exceptions.ClientResponseError: 403, message='Forbidden',
#       409 when trying to stop a time entry that's already stopped... etc
# When sending a request BODY with verb GET (or just a bad request in general)
#   aiohttp.client_exceptions.ClientResponseError: 400, message='Bad Request',
##
# TODO: can probably re-factor a bit and remove the do_$VERB_request functions and replace with
#   a singular do_request which takes an aiohttp.request() object



"""
Constants for the Toggl API.
Implements a subset of the complete API.

"""

from . import __version__ as version

BASE = "https://api.track.toggl.com/api/v9"

CURRENT_RUNNING_TIME = f"{BASE}/time_entries/current"
PROJECTS = f"{BASE}/projects"
START_TIME = f"{BASE}/time_entries/start"

DEFAULT_CREATED_BY = "lib-toggl"
USER_AGENT = f"{DEFAULT_CREATED_BY} ({version})"



"""Basic tests of the client module"""

import pytest

from lib_toggl.client import Toggl


@pytest.mark.asyncio
async def test_client_init_no_api_key():
    """Tests that client requires an API key"""
    with pytest.raises(TypeError):
        # pylint: disable=no-value-for-parameter
        _ = Toggl()  # pyright: ignore reportCallIssue


@pytest.mark.asyncio
async def test_client_init_with_api_key():
    """Tests that client requires an API key"""
    _key = "fake_api_key"
    x = Toggl(_key)
    assert x.api_key == _key


##
# Most of client is boilerplate for API calls and I'm not in the mood to mock aiohttp since the data
#   is almost always directly piped into Pydantic models.
# The update_tags() function does have some non-standard logic that's worth testing.
# But it will require mocking out some of the API calls.
# Let's call this a TODO for now.
##



#!/usr/bin/env python3
"""
Dump tag name/id pairs to a file for reference
"""

import asyncio
import json
import logging
import os
import sys

from lib_toggl.client import Toggl

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


async def main():
    """Does the needful"""
    _key = os.getenv("TOGGL_API_KEY")
    async with Toggl(_key) as api:
        w = await api.get_workspaces()
        if not w:
            log.error("No workspaces")
            sys.exit(1)

        workspace_id = w[0].id
        if workspace_id is None:
            log.error("No workspace id found")
            sys.exit(1)

        log.info("Using workspace id %s", workspace_id)
        tags = await api.get_tags(workspace_id)
        tag_map = {tag.name: tag.id for tag in tags}
        with open("tag_map.json", "w", encoding="utf8") as json_file:
            # Pretty print
            json.dump(tag_map, json_file, indent=4, sort_keys=True)

    sys.exit()


if __name__ == "__main__":
    asyncio.run(main())



"""
Test the version of the package.
"""

import re


def test_version_found():
    """Assuming importlib.metadata.version() works, confirm version is a string."""
    from lib_toggl import __version__  # pylint: disable=import-outside-toplevel

    assert __version__ is not None
    assert re.match(r"\d+\.\d+\.\d+", __version__)



"""Represents a Toggl Tag object."""

import logging
from datetime import datetime
from typing import Optional

try:
    # Pydantic v2 ships a copy of v1.
    from pydantic.v1 import BaseModel, Field
except ImportError:
    # Home Assistant does not yet support v2.
    from pydantic import BaseModel, Field

from .const import BASE

log = logging.getLogger(__name__)


@staticmethod
# pylint: disable=invalid-name
def TAGS_ENDPOINT(workspace_id: int | None) -> str:
    """Returns the endpoint for managing Tags in a particular worksapce."""
    if not workspace_id:
        raise ValueError("workspace_id must be specified")
    return f"{BASE}/workspaces/{workspace_id}/tags"


class Tag(BaseModel):  # pyright: ignore[reportGeneralTypeIssues]
    """Class representing Tag object.
    Leverages dataclass to cut down on boilerplate code.
    See: https://engineering.toggl.com/docs/api/tags
    """

    # Name of the tag
    name: str = Field(default=None, description="Tag Name, required.")

    # Optional to account for user creating a Tag object to send to API
    id: Optional[int] = Field(default=None, description="Tag ID.")

    workspace_id: int = Field(
        description="Workspace ID tag is associated with, required.", default=None
    )

    creator_id: Optional[int] = Field(
        description="ID of the user who created the tag, optional.", default=None
    )

    at: Optional[datetime] = Field(
        exclude=True, default=None, description="When Tag was last updated", repr=False
    )

    deleted_at: Optional[datetime] = Field(
        exclude=True, default=None, description="When was deleted, null if not deleted"
    )

    # Dev docs don't explain what this is?
    permissions: Optional[str] = Field(
        default=None, exclude=True, repr=False, description="permissions"
    )



#!/usr/bin/env python3
"""
Super simple PoC for playing with the API Client
"""

import asyncio
import logging
import os
import sys
from datetime import datetime
from time import sleep

from lib_toggl.client import Toggl
from lib_toggl.time_entries import TimeEntry

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)

# Time to pause between steps to allow for manual verification
SLEEP_TIME_SECONDS = 5


async def main():
    """Does the needful"""
    _key = os.getenv("TOGGL_API_KEY")
    async with Toggl(_key) as api:
        # Start with getting account information.
        # This does two things: 1) confirms API key is valid, 2) gets user ID
        ##
        # Can also do get_account_details()
        me = await api.account
        log.info("account details are:", extra={"account": me})

        log.info("Fetching workspaces")
        w = await api.get_workspaces()
        log.info("Got workspace!", extra={"count": len(w), "workspaces": w})

        # I am not a premium user, so I only have one workspace.
        if len(w) > 1:
            log.warning("More than one workspace found, using first one")

        workspace_id = w[0].id
        log.info("Using workspace id %s", workspace_id)

        log.info("Fetching Current Time Entry")
        te = await api.current_time_entry
        if te is None:
            log.info("No current time entry")
        else:
            log.info("get_current_time_entry", extra={"entry": te})

        log.info("Creating a new Time Entry")

        body = {
            "description": "Testing from `lib-toggl/poc.py`!",
            "tags": ["test-tag"],
            "workspace_id": workspace_id,
        }

        new_time_entry = TimeEntry(**body)
        log.debug("new_time_entry: %s", new_time_entry)
        created_te = await api.create_new_time_entry(new_time_entry)
        log.info("created_te: %s", created_te)
        if created_te is None:
            log.error("Failed to create time entry")
            sys.exit(1)

        log.info(
            "Sleeping %s s before editing time entry to add tags...",
            SLEEP_TIME_SECONDS,
        )
        sleep(SLEEP_TIME_SECONDS)

        # This will REMOVE the `test-tag` from the time entry and replace it with `tag-that-might-not-exist`.
        # As the name implies, the tag might not exist. It will be created if it does not.
        created_te.tags = ["tag-that-might-not-exist"]
        # And just for giggles, update the description, too.
        created_te.description = "Updated from `lib-toggl/poc.py`!"
        # This will result in a few API calls to create tag if needed and then update the time entry tags and description.
        correct_te = await api.edit_time_entry(created_te)
        log.info("correct_te: %s", correct_te)

        log.info(
            "Sleeping %s s before editing time entry to clear tags...",
            SLEEP_TIME_SECONDS,
        )
        sleep(SLEEP_TIME_SECONDS)

        created_te.tags = []
        created_te.description = "Updated from `lib-toggl/poc.py` with no tags!"
        correct_te = await api.edit_time_entry(created_te)
        log.info("correct_te: %s", correct_te)

        log.info(
            "Sleeping %s s before stopping",
            SLEEP_TIME_SECONDS,
        )
        sleep(SLEEP_TIME_SECONDS)

        result = await api.stop_time_entry(created_te)
        log.info("stop_time_entry", extra={"result": result})

        log.info("Fetching time entries")
        _now = datetime.now()
        _today = datetime(_now.year, _now.month, _now.day)
        te = await api.get_time_entries(start_date=_today, end_date=_now)
        log.info("get_time_entries", extra={"count": len(te), "entries": te})

    sys.exit()


if __name__ == "__main__":
    asyncio.run(main())



"""Various type/class definitions for the Organizations bit of the Toggle API
Parse/Coercion done by Pydantic
"""

try:
    # Pydantic v2 ships a copy of v1.
    from pydantic.v1 import BaseModel
except ImportError:
    # Home Assistant does not yet support v2.
    from pydantic import BaseModel


from .const import BASE

ENDPOINT = f"{BASE}/organizations"


class Organization(BaseModel):  # pyright: ignore[reportGeneralTypeIssues]
    """Class representing the Toggl organization object.
    Leverages dataclass to cut down on boilerplate code.
    See: https://developers.track.toggl.com/docs/api/workspaces#200
    """
